function [ normval, fopt, info ] = linorm( E, A, B, C, D, options )
%
% PURPOSE:
%
% To compute H-infinity-norm of an L-infinity transfer function
%
%   G(s) = C*(sE - A)^(-1)*B + D
%
% corresponding to a continuous-time linear time-invariant descriptor 
% system ( sE-A, B, C, D ). The matrices E and A are assumed to be
% sparse. The matrix pencil sE-A is assumed to have no purely imaginary
% eigenvalues. It is also assumed that the number of inputs and ouputs is 
% small compared to the system order.
%
% The L-infinity-norm is defined as
%
%   ||G||_Linf := sup_{w real} ||G(iw)||_2.
%
% ARGUMENTS:
%
% Inputs:
%
% E, A, B, C, D: System matrices defining the transfer function.
% options      : Options struct (use default values if empty). The 
%                following options can be specified:
%    options.tol     : desired accuracy of final L-infinity-norm value
%                      (default = 1e-6);
%    options.inftol  : tolerance for w for evaluating 
%                      lim_{w -> infinity} ||G(iw)||_2 (default = 1e-8);
%    options.domtol  : tolerance for deciding which dominant poles get 
%                      used as shifts for calculating the initial 
%                      eigenvalues of the even matrix pencil
%                      (default = 0.5);
%    options.npolinit: number of dominant poles computed in initial stage
%                      (default = 20);
%    options.shift   : initial shift for dominant pole algorithm
%                      (default = 1i);
%    options.shifttol: if 2 or more shifts are separated by less than the 
%                      product of shifttol with the largest dominant pole, 
%                      they are averaged together before being used as an 
%                      input to even IRA (default = 0.01);
%    options.prtlevel: specifies the print level as follows:
%                      = 0: return no information;
%                      = 1: return minimal information (warnings, 
%                           intermediate gamma values);
%                      = 2: return most information (+ imaginary 
%                           eigenvalues found);
%                      = 3: all information (+ dominant poles);
%                      (default = 0);
%    options.mmax    : maximum search space dimension for even IRA
%                      (default = 8);
%    options.maxcycle: maximum number of restarts for even IRA
%                      (default = 30).
%    options.nwanted : number of desired eigenvalues calculated per shift 
%                      in even IRA (default = 4).
%    options.eigtol  : eigenvalue tolerance for even IRA (default = 1e-6).
%    options.sweeptol: if only two remaining eigenvalues are within 
%                      this tolerance, a frequency sweep is performed to 
%                      calculate the norm (default = 1e-5).
%    options.divnum  : number of points at which the singular values are
%                      calculated for the frequency sweep (default = 100).
%    options.condzero: where the transfer function G is evaluated (instead
%                      of zero) if cond(A) = Inf (default = 1e-6).
%    options.movechk : for displacing the shifts, the amount the routine
%                      checks on either side of the shift in order to 
%                      choose which direction to move it (default = 1e-4).
%    options.shiftmv : (relative) amount by which the routine displaces 
%                      the shifts (default = 5e-4).
%    
%
% Outputs:
%
% normval      : Computed L-infinity-norm.
% fopt         : Frequency at which the norm value is attained.
% info         : Info struct that contains the following information:
%    info.error      : contains an error indicator as follows:
%                      = 0: return without error;
%                      = 1: computed H-infinity-norm smaller than 
%                           ||G(iw_0)||_2 for a test frequency w_0, or
%                           smaller than previously calculated gamma_lb. 
%                      = 2: computed eigenvalues are too close together to
%                           be properly resolved.
%    info.cycle      : number of iterations;
%    info.time       : time needed to compute result.
%
% REFERENCES:
%
% [1] P. Benner and M. Voigt:
%     L-infinity-norm Computation for Continuous-Time Descriptor Systems
%     Using Structured Matrix Pencils,
%     IEEE Trans. Automat. Control, 57(1):233-238, 2012.
%
% [2] P. Benner and M. Voigt:
%     A structured pseudospectral method for H-infinity-norm computation of 
%     large-scale descriptor systems, 2012.
%
% AUTHORS:
%
% Ryan Lowe, Max Planck Institute for Dynamics of Complex Technical
% Systems, Magdeburg, Germany, 08/14/13.
%
% Matthias Voigt, Max Planck Institute for Dynamics of Complex Technical
% Systems, Magdeburg, Germany.
%
% REVISIONS:
%
% -
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  LINORM 1.0 Copyright (C) 2013 Ryan Lowe, Matthias Voigt 
%
%  This program is free software: you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation, either version 3 of the License, or
%  (at your option) any later version.
%
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License for more details.
%
%  You should have received a copy of the GNU General Public License
%  along with this program.  If not, see <http://www.gnu.org/licenses/>.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%% CHECK ARGUMENTS.
%
tic;
%
% Check the number of arguments.
%
if ( nargin < 5 )
    error( 'LINORM requires at least 5 input arguments.' );
elseif ( nargin == 5 )
    options = [];
elseif ( nargin > 6 )
    error( 'LINORM requires at most 6 input arguments.' );
end
%
if ( nargout > 3 )
    error( 'LINORM requires at most 3 output arguments.' );
end
%
% Check dimensions.
%
[ n1, p1 ] = size( B );
[ m1, n2 ] = size( C );
[ m2, p2 ] = size( D );
%
if ( n1 ~= n2 || m1 ~= m2 || p1 ~= p2 )
    error( 'The dimensions of B, C, and D are not compatible.' );
else
    p = p1;
    m = m1;
end
%
[ ne1, ne2 ] = size( E );
[ na1, na2 ] = size( A );
if ( ne1 ~= ne2 )
    error( 'The matrix E must be square.' );
elseif ( na1 ~= na2 )
    error( 'The matrix A must be square.' );
elseif ( ne1 ~= na1 )
    error( 'The dimensions of E and A are not compatible.' );
elseif ( ne1 ~= n1 )
    error( 'The dimensions of E and A are not compatible with B, C, and D.' );
else
    n = n1;
end
%
% Check for sparsity.
%
if ( ~issparse( E ) )
    error( 'The matrix E must be sparse.' );
elseif ( ~issparse( A ) )
    error( 'The matrix A must be sparse.' );
end
%
%% SET OPTIONS.
%
% tol: desired accuracy of final L-infinity-norm value.
%
if isfield( options, 'tol' )
    tol = options.tolout; 
else % default
    tol = 1e-6;
end
%
% inftol: tolerance for w for evaluating lim_{w -> infinity} ||G(iw)||_2.
%
if isfield( options, 'inftol' )
    inftol = options.inftol; 
else % default
    inftol = 1e-8;
end
%
% domtol: tolerance for deciding which dominant poles get used as shifts
% for calculating the initial eigenvalues of the even matrix pencil.
%
if isfield( options, 'domtol' )
    domtol = options.domtol; 
else % default
    domtol = 0.5;
end
%
% npolinit: number of dominant poles computed in initial stage.
%
if isfield( options, 'npolinit' )
    npolinit = options.npolinit; 
else % default
    npolinit = 20;
end
%
% shift: initial shift for dominant pole algorithm.
%
if isfield( options, 'shift' )
    shift = options.shift; 
else % default
    shift = 1i;
end
%
% shifttol: if 2 or more shifts are separated by less than the product of
% shifttol with the largest dominant pole, they are averaged together 
% before being used as an input to even IRA.
%
if isfield( options, 'shifttol' )
    shifttol = options.shifttol; 
else % default
    shifttol = 0.01;
end
%
% prtlevel: specifies the print level.
%
if isfield( options, 'prtlevel' )
    prtlevel = options.prtlevel;
    if ( prtlevel ~= 0 && prtlevel ~= 1 && prtlevel ~= 2 && prtlevel ~= 3 )
         error( 'The argument options.prtlevel must be from 0 to 3.' );
    end
else % default
    prtlevel = 0;
end
%
% mmax: maximum search space dimension for even IRA.
%
if isfield( options, 'mmax' )
    mmax = options.mmax; 
else % default
    mmax = 8;
end
%
% maxcycle: maximum number of restarts for even IRA.
%
if isfield( options, 'maxcycle' )
    maxcycle = options.maxcycle; 
else % default
    maxcycle = 30;
end
%
% eigtol: eigenvalue tolerance for even IRA.
%
if isfield( options, 'eigtol' )
    eigtol = options.eigtol; 
else % default
    eigtol = 1e-6;
end
%
% nwanted: number of desired eigenvalues calculated per shift in even IRA.
%
if isfield( options, 'nwanted' )
    nwanted = options.nwanted; 
else % default
    nwanted = 4;
end
%
% sweeptol: if only two remaining eigenvalues are within this tolerance, a
% frequency sweep is performed to calculate the norm. 
%
if isfield( options, 'sweeptol' )
    sweeptol = options.sweeptol; 
else % default
    sweeptol = 1e-5;
end
%
% divnum: number of points at which the singular values are calculated for 
% the frequency sweep
%
if isfield( options, 'divnum' )
    divnum = options.divnum; 
else % default
    divnum = 100;
end
%
% condzero: where the transfer function G is evaluated (instead of zero) if
% cond(A) = Inf.
%
if isfield( options, 'condzero' )
    condzero = options.condzero; 
else % default
    condzero = 1e-6;
end
%
% movechk: for displacing the shifts, the amount the routine checks on 
% either side of the shift in order to choose which direction to move it.
%
if isfield( options, 'movechk' )
    movechk = options.movechk; 
else % default
    movechk = 1e-4;
end
%
% shiftmv: (relative) amount by which the routine displaces the shifts.
%
if isfield( options, 'shiftmv' )
    shiftmv = options.shiftmv; 
else % default
    shiftmv = 5e-4;
end
%% COMPUTATIONS.
%
warning off all;
%
% Compute test frequencies. Set options for SAMDP.
%
optsamdp.yEx_scaling = 0;
optsamdp.strategy = 'LR';
optsamdp.nwanted = npolinit;
optsamdp.displ = 0;
info.error = 0;
%
[ poles, ~, rightev, leftev, ~ ] = samdp( A, E, B, C', D, shift, ...
    optsamdp );
%
% Extract the poles with non-negative imaginary part.
%
posimag = find( imag( poles ) >= 0 );
poles = poles( posimag );
rightev = rightev( :, posimag );
leftev = leftev( :, posimag );
ndpol = numel( rightev( 1, : ) );
%
% Scale the eigenvectors such that y*Ex = 1.
%
for j = 1 : ndpol
    tmp = leftev( :, j )'*E*rightev( :, j );
    tmp = tmp/abs( tmp );
    rightev( :, j ) = rightev( :, j )*sqrt( tmp );
    leftev( :, j ) = leftev( :, j )*sqrt( tmp );
end
%
% Sort the computed dominant poles and the associated eigenvectors with
% respect to the largest |residue|/|real(pole)| magnitude.
%
contobs = zeros( 1, ndpol );
for j = 1 : ndpol 
    contobs( j ) = norm( C*rightev( :, j )*leftev( :, j )'*B, 2 )/ ...
        abs( real( poles( j ) ) );
end
[ dom, pos ] = sort( contobs, 'descend' );
poles = poles( pos );
%
% Evaluate lim_{w -> infinity} ||G(iw)||_2.
%
omega = abs( max( imag( poles ) ) );
ressum = Inf;
while ( ressum > inftol )
    omega = 10*omega;
    ressum = 0;
    for j = 1 : ndpol
        ressum = ressum + dom( j )/( omega - imag( poles( j ) ) );
    end
end
svinf = max( svds( C/( 1i*omega*E - A )*B + D ) );
%
% Evaluate ||G(iw_0)||_2 at test frequencies w_0.
%
svmax = 0;
for j = 1 : ndpol
    sv = max( svds( C/( 1i*imag( poles( j ) )*E - A )*B + D ) );
    if ( sv > svmax )
        svmax = sv;
        polenum = j;
    end
end
%
polesort = sort( imag( poles ) );
%
% Calculate the maximum singular value of G(0). 
%
if ( cond( A ) < 1/eps )
    svzero = max( svds( C/( -A )*B + D ) );
else
    svzero = max( svds( C/( condzero*E - A )*B + D ) );
end
%
% Set gamma_lb to be the maximum singular value of G evaluated at 0, the
% dominant poles, and at infinity.
%
maxsing( 1 ) = svzero;
maxsing( 2 ) = svinf;
maxsing( 3 ) = svmax;
gamma_lb = max( maxsing );
%
if ( gamma_lb == svzero )
    fopt = 0;
elseif ( gamma_lb == svinf )
    fopt = Inf;
else
    fopt = imag( poles( polenum ) );
end
%
% Transfer all poles with magnitude greater than the product of domtol
% and the magnitude of the most dominant pole, into the shift matrix used
% to calculate the eigenvalues of the even matrix pencil.
%
domcount = 0;
for k = 1 : size( poles )
    if ( dom( k ) > domtol*dom( 1 ) )
        shift( k ) = poles( k );
        domcount = domcount + 1;
    end
end
%
% Print the most dominant poles to the user, if desired.
%
if ( prtlevel > 2 )
    fprintf( 'The %i most dominant poles with associated dominance values are:\n', domcount );
    for k = 1 : domcount
        fprintf( '%.3f + %.3fi with dominance %f\n', real( poles( k ) ), imag( poles( k ) ), dom( k ) );
    end
    fprintf( '\n' );
end
if ( prtlevel > 1 )
    fprintf( 'The %i most dominant poles are used to calculate the initial gamma value.\n\n', domcount );
end
%
% Construct the even matrix pencil lambda*N - M, where
%
%                  [      0       -lamdba*E' - A' |    -C'        0     ]
%                  [ lambda*E - A        0        |     0        -B     ]
%  lambda*N - M =  [----------------------------------------------------].
%                  [     -C              0        | gamma*I_p    -D     ]
%                  [      0             -B'       |    -D'    gamma*I_m ]
%
% N consists of the linear part.
% 
N = [ sparse( n, n ), -E', sparse( n, m + p ); ...
    E, sparse( n, n + m + p ); ...
    sparse( m + p, 2*n + m + p ) ];
%
% M consists of the constant part.
% 
M = [ sparse( n, n ), A', C', sparse( n, p ); ...
    A, sparse( n, n + m ), B; ...
    C, sparse( m, n + m ), D; ...
    sparse( p, n ), B', D', sparse( p, p ) ];
%
% Set the options for even IRA, stored in optseig.
%
optseig.tol = eigtol; 
optseig.dimmax = mmax;
optseig.dimrestart = nwanted/2;
optseig.nrestart = maxcycle;
optseig.powerfirst = 3;
optseig.buildNV = 1;
optseig.OPform = 2;
optseig.disp = 0;
%
cycle = 0;
while ( true )
    cycle = cycle + 1;
    %    
    % Set the value of gamma.
    %    
    gamma = ( 1 + 2*tol )*gamma_lb;
    %
    % Update the bottom right block of M with new gamma value.
    %
    M( 2*n + 1 : 2*n + m, 2*n + 1 : 2*n + m ) = gamma*speye( m, m );
    M( 2*n + m + 1 : end, 2*n + m + 1 : end ) = gamma*speye( p, p );
    %
    % If the shift is too close to one of the eigenvalues produced by even 
    % IRA algorithm, all other eigenvalues may be inaccurate. Thus, the 
    % shifts are offset by a small amount in the direction away from this 
    % eigenvalue.
    %
    for k = 1 : length( shift )
        svlow = max( svds( C/( shift( k )*( 1 - movechk )*E - A )*B + ...
            D ) );
        svhigh = max( svds( C/( shift( k )*( 1 + movechk )*E - A )*B ... 
            + D ) );
        if ( svlow < svhigh )
            shift( k ) = shift( k )*( 1 - shiftmv );
        else
            shift( k ) = shift( k )*( 1 + shiftmv );
        end    
    end
    %
    % Compute the eigenvalues of the matrix pencil lambda*N - M using the 
    % even IRA algorithm. The shifts used to call the function are the
    % imaginary parts of the dominant poles previously calculated (for the
    % first iteration), and the midpoint of the intervals given by the 
    % eigenvalues with zero real part (for every further iteration).
    %
    % To reduce the number of calls to even IRA, all imaginary parts of
    % the shifts that are close enough together are combined into a single 
    % shift.
    %
    imagshift = sort( imag( shift ) );
    for k = 1 : length( imagshift ) - 1
        if ( imagshift( k ) ~= 0 )
            j = k;
            %
            % Adjacent shifts within shifttol of eachother are averaged, 
            % and the previous values are eliminated.            
            %
            while ( abs( imagshift( k ) - imagshift( j ) ) <= shifttol* ...
                    abs( polesort( end ) ) )
                j = j + 1;
                if ( j == length( imagshift ) + 1 )
                    break;
                end
            end
            j = j - 1;
            if ( j ~= k )
                imagshift( k ) = 0.5*( imagshift( j ) + imagshift( k ) );
                imagshift( k : j - 1 ) = zeros( j - k , 1 );
            end
        end
    end
    %
    % Remove the zeros from the imagshift vector.
    %
    imagshift = imagshift( imagshift ~= 0 );    
    %
    % Even IRA call loop.
    %
    eigen = zeros( length( imagshift ), nwanted + 2 );
    for j = 1 : length( imagshift )
        %
        % eigentemp is new array output with list of found eigenvalues for
        % each shift.
        %
        v1 = rand( size( M, 1 ), 1 );
        v1 = v1/norm( v1 );
        optseig.v1 = v1;       
        eigvals = even_ira( M, N, 1i*imagshift( j ), nwanted, optseig );
        eigen( j, 1 : length( eigvals ) ) = eigvals;
    end
    %
    % We now check if there are any purely imaginary eigenvalues of the
    % even matrix pencil lambda*N - M. Introduce eigen_imag, which will  
    % contain all purely imaginary eigenvalues of the even matrix pencil 
    % lambda*N - M.
    %
    ind_eigen = find( eigen ~= 0 );
    ind_imag = find( real( eigen( ind_eigen ) ) == 0 );
    eigen_imag = imag( eigen( ind_imag ) );
    %
    % Print information about gamma bound to the user, if desired.        
    %
    if ( cycle == 1 )
        if ( prtlevel > 0 )
            fprintf( 'The initial lower bound of gamma is %f at fopt = %f.\n', gamma_lb, fopt );              
            fprintf( '\nFor cycle 1:\n' );      
        end
    else
        if ( prtlevel > 0 )
            fprintf( 'The lower bound of gamma is %f at fopt = %f.\n', gamma_lb, fopt );              
            fprintf( '\nFor cycle %i:\n', cycle );      
        end
    end
    shiftcount = 0;
    if ( prtlevel > 1 )
        %
        % Print the shifts and the associated eigenvalues to the user, if 
        % desired.
        %
        for j = 1 : length( imagshift )
            %
            % For a given shift, checks if there are any eigenvalues with 0
            % real part.
            %
            isimag = 0;
            for k = 1 : length( eigen( j , : ) )
                if ( eigen( j, k ) ~= 0 && real( eigen( j, k ) ) == 0 )
                    isimag = 1;
                end
            end
            eigen( j, : ) = sort( eigen( j, : ) );          
            if ( isimag == 1 )
                shiftcount = shiftcount + 1;
                fprintf( 'The positive imaginary eigenvalues for shift = %.9fi are: \n', imagshift( j ) );
                for k = 1 : length( eigen( j, : ) )
                    %
                    % Displays only positive imaginary eigenvalues.
                    %
                    if ( real( eigen( j, k ) ) == 0 && ...
                            imag( eigen( j, k ) ) > 0 )
                        fprintf( 'lambda = %.9fi\n', imag( eigen( j, k ) ) );
                    end
                end
            end
        end
        fprintf( 'There are %i shifts, and %i shifts that produce imaginary eigenvalues.\n', length( imagshift ), shiftcount );              
    end
    if ( prtlevel > 0 )
        fprintf( 'There are %i eigenvalues, and %i imaginary eigenvalues.\n', length( ind_eigen ), length( ind_imag ) );               
    end
    %
    % Sort eigen_imag, while removing all negative eigenvalues.
    %
    eigen_imag = sort( eigen_imag( eigen_imag > 0 ) );
    %
    % If there are no purely imaginary eigenvalues, then 
    % gamma > L-infinity-norm, so the norm value must lie in the interval
    % ( gamma_lb, gamma ).
    %
    if ( isempty( eigen_imag ) )
        gamma_ub = gamma;
        break;
    else
        %
        % If there are purely imaginary eigenvalues, we evaluate the
        % transfer function at the midpoints between these eigenvalues, and
        % then take the largest singular value as the next gamma_lb.
        %
        % Check whether a frequency sweep is necessary.
        %
        if ( length( eigen_imag ) == 2 && ...
            abs( ( eigen_imag( 1 ) - eigen_imag( 2 ) )/ ...
            eigen_imag( 1 ) ) < sweeptol )
            if ( prtlevel > 0 )
                fprintf( '\nWarning: The final eigenvalues are too close together.\n' );
                fprintf( 'The final norm value will be determined using a frequency sweep, and\n' );
                fprintf( 'is not guaranteed to be within the desired tolerance.\n' );
            end
            %
            % Sweep through the singular values around the final 2
            % eigenvalues. This has been (probably unnecessarily)
            % generalized for more than 2 eigenvalues. 
            %
            numpeaks = length( eigen_imag )/2;
            normsweep = zeros( divnum*numpeaks, 1 );
            for j = 1 : numpeaks
                diff = eigen_imag( 2*j ) - eigen_imag( 2*j -1 );
                for k = 1 : divnum
                    normsweep( ( j - 1 )*numpeaks + k ) = max( svds( C/ ...
                        ( 1i*( eigen_imag( 2*j - 1 ) + 5*k/divnum* ...
                        diff - 2*diff )*E - A )*B + D ) );
                end
            end
            [ gamma_lb, index ] = max( normsweep );
            gamma_ub = gamma_lb;
            peakindex = floor( ( index - 1 )/divnum ) + 1;
            peakdiff = eigen_imag( 2*peakindex ) - ... 
                eigen_imag( 2*peakindex -1 );
            fopt = eigen_imag( 2*peakindex - 1 ) + peakdiff*( index - ... 
                ( peakindex - 1 )*divnum )/divnum;
            info.error = 2;
            break;
        end
        %
        % Calculate the midpoints of the eigenvalues.
        %
        mid = zeros( length( eigen_imag ) - 1, 1 );
        for j = 1 : length( eigen_imag ) - 1
            mid( j ) = sqrt( eigen_imag( j )*eigen_imag( j + 1 ) );
        end
        %
        % Find the singular values of the transfer function at these
        % midpoints.
        %
        maxsing = zeros( length( mid ), 1 );
        for j = 1 : length( mid )
            maxsing( j ) = max( svds( C/( 1i*mid( j )*E - A )*B + D ) );
        end
        %
        % gamma_lb is set to the maximum singular value over all of the
        % midpoints.
        %
        prevgamma = gamma_lb;
        [ gamma_lb, index ] = max( maxsing );
        %
        % If the new gamma_lb value is less than the one calculated for the
        % previous cycle, an error has occurred.
        %
        if ( prevgamma > gamma_lb )
            info.error = 1;
            if ( prtlevel > 0 )
                fprintf( '\nWarning: New gamma value is less than the gamma value for the\n' );
                fprintf( 'previous cycle. An error may have occured.\n' );
                fprintf( 'The final norm value is not guaranteed to be within desired tolerance\n' ); 
                fprintf( 'of actual value.\n' );
            end
        end
        %
        % Assign value for the fopt, and set the next shift values as the
        % previous midpoints.
        %
        fopt = mid( index );
        shift = 1i*mid( index );
    end       
end
%
% If no imaginary eigenvalues are found in the first cycle, print out a
% warning message to the user.
%
if ( cycle == 1 && info.error ~= 2 && fopt ~= 0 && fopt ~= Inf && ...
        prtlevel > 0 )
    fprintf( '\nWarning: No imaginary eigenvalues found in first cycle.\n' );
    fprintf( 'The tolerance on the eigenvalues given by eigtol might be too large.\n' );
end
%
% Calculate the final L-infinity-norm.
%
normval = 0.5*( gamma_lb + gamma_ub );
%
% Print the final calculated L-infinity-norm value and code execution time 
% to the user, if desired.
%
time = toc;
if( prtlevel > 0 )
    fprintf( '\nThe L-infinity-norm is %g at fopt = %g.\n', normval, fopt );
    fprintf( 'The runtime is %g seconds.\n', time );
end
%
% Return further information to the user in struct 'info'.
%
info.cycle = cycle;
info.time = time;
if ( normval < svinf || normval < svzero || normval < svmax )
    info.error = 1;
end
%
return